import re
from inspect import isclass, isfunction
from typing import Callable, Dict, Iterable, Optional, Union

from django.utils.module_loading import import_string
from django.views.generic import View

from .logging import logger

name_regex = re.compile(r'[-\w.]+')


class LazyView:
    __slots__ = ('path', 'initkwargs', 'view_func')

    def __init__(self, view_path: str, **initkwargs):
        self.path = view_path
        self.initkwargs = initkwargs
        self.view_func = None  # type: Optional[Callable]

    def __getattr__(self, item):
        if self.view_func is None:
            self.view_func = self._resolve()
            logger.debug('View "%s" instantiated.' % self.path)
        return getattr(self.view_func, item)

    def __call__(self, *args, **kwargs):
        if self.view_func is None:
            self.view_func = self._resolve()
            logger.debug('View "%s" instantiated.' % self.path)
        return self.view_func(*args, **kwargs)

    def _resolve(self) -> Callable:
        view = import_string(self.path)
        if isfunction(view):
            return view
        elif isclass(view) and issubclass(view, View):
            return view.as_view(**self.initkwargs)
        else:
            raise TypeError(view)


class Registry:
    __slots__ = ('_registry',)

    def __init__(self):
        self._registry = {}  # type: Dict[str, Callable]

    def __iter__(self):
        return iter(self._registry)

    def __getitem__(self, item):
        return self._registry[item]

    def register(
        self, names: Union[str, Iterable], view: Union[Callable, View], **initkwargs
    ):
        """
        :type names: str | list | tuple
        :type view: types.FunctionType | django.views.generic.View
        """
        if isinstance(names, str):
            names = [names]

        # check names
        for name in names:
            if not name_regex.fullmatch(name):
                raise ValueError(
                    "Invalid view name: '{}'. Names must contain only letters, "
                    "numbers, dots (.), dashes (-) and underscores (_)".format(name)
                )
            if name in self._registry:
                logger.warning('view `%s` was already registered.' % name)

        view_path = '.'.join((view.__module__, view.__qualname__))
        lazy_view = LazyView(view_path, **initkwargs)

        # assigning by reference
        for name in names:
            self._registry[name] = lazy_view


registry = Registry()



import django

from .registry import registry
from .views import router

if django.VERSION >= (2, 2):  # noqa
    from django.urls import ResolverMatch, URLPattern
    from django.urls.resolvers import RegexPattern

    class SSIURLPattern(URLPattern):
        def resolve(self, path):
            match = self.pattern.match(path)
            if match:
                new_path, args, kwargs = match
                name = kwargs.pop('name')
                if name not in registry:
                    return
                view = registry[name]
                return ResolverMatch(
                    view, args, kwargs, self.pattern.name, route=str(self.pattern)
                )

    def ssi_url(regex, view, kwargs=None, name=None):
        pattern = RegexPattern(regex, name=name, is_endpoint=True)
        return SSIURLPattern(pattern, view, kwargs, name)


elif django.VERSION >= (2, 0):
    from django.urls import ResolverMatch, URLPattern
    from django.urls.resolvers import RegexPattern

    class SSIURLPattern(URLPattern):  # type: ignore
        def resolve(self, path):
            match = self.pattern.match(path)
            if match:
                new_path, args, kwargs = match
                name = kwargs.pop('name')
                if name not in registry:
                    return
                view = registry[name]
                return ResolverMatch(view, args, kwargs, self.pattern.name)

    def ssi_url(regex, view, kwargs=None, name=None):
        pattern = RegexPattern(regex, name=name, is_endpoint=True)
        return SSIURLPattern(pattern, view, kwargs, name)


else:
    from django.core.urlresolvers import RegexURLPattern, ResolverMatch

    class SSIURLPattern(RegexURLPattern):  # type: ignore
        def resolve(self, path):
            match = self.regex.search(path)
            if match:
                kwargs = match.groupdict()
                name = kwargs.pop('name')
                if name not in registry:
                    return
                view = registry[name]
                return ResolverMatch(view, (), kwargs, self.name)

    def ssi_url(regex, view, kwargs=None, name=None):
        return SSIURLPattern(regex, view, kwargs, name)


app_name = 'ssi_views'
urlpatterns = [
    ssi_url(r'(?P<name>[-\w.]+)/', router, name='router'),
]



from django.http import HttpResponseNotFound

from .logging import logger
from .registry import registry


def router(request, name: str, *args, **kwargs):
    try:
        view = registry[name]
    except KeyError:
        logger.debug('Resolving "%s"...' % name)
        return HttpResponseNotFound()
    else:
        return view(request, *args, **kwargs)



from django.template import Library
from django.utils.html import format_html

from ..logging import logger
from ..registry import registry

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

try:
    import jinja2
except ImportError:
    jinja2 = None


register = Library()


@register.simple_tag(name='ssi_url')
def do_ssi_url(name):
    if name not in registry:
        logger.warning('view `%s` is not registered' % name)
    return reverse('ssi_views:router', kwargs={'name': name})


@register.simple_tag(name='ssi_include')
def do_ssi_include(name):
    if name not in registry:
        logger.warning('view `%s` is not registered' % name)
    url = reverse('ssi_views:router', kwargs={'name': name})
    return format_html('<!--# include virtual="{}" -->', url)


if jinja2 is not None:
    from jinja2 import nodes
    from jinja2.ext import Extension

    class SSIIncludeExtension(Extension):
        tags = {'ssi_include'}

        def parse(self, parser):
            lineno = next(parser.stream).lineno
            view_name = parser.parse_expression()
            call = self.call_method('_ssi_include', [view_name])
            return nodes.Output([call], lineno=lineno)

        @staticmethod
        def _ssi_include(view_name):
            return do_ssi_include(view_name)

    # django-jinja support
    try:
        from django_jinja import library
    except ImportError:
        pass
    else:
        library.global_function(name='ssi_url')(do_ssi_url)
        library.extension(SSIIncludeExtension)
